@@ -24,8 +24,6 @@
 
 #include <caml/osdeps.h>
 
-#define INLINE __inline
-
 #if 1
 #define TRACE(x)
 #define TRACE1(x,y)
@@ -40,18 +38,37 @@ typedef DWORD st_retcode;
 
 #define SIGPREEMPTION SIGTERM
 
+/* Unique thread identifiers and atomic operations over them */
+#ifdef ARCH_SIXTYFOUR
+typedef LONG64 st_tid;
+#define Tid_Atomic_Exchange InterlockedExchange64
+#define Tid_Atomic_Compare_Exchange InterlockedCompareExchange64
+#else
+typedef LONG st_tid;
+#define Tid_Atomic_Exchange InterlockedExchange
+#define Tid_Atomic_Compare_Exchange InterlockedCompareExchange
+#endif
+
 /* Thread-local storage associating a Win32 event to every thread. */
 static DWORD st_thread_sem_key;
 
+/* Thread-local storage for the OCaml thread ID. */
+static DWORD st_thread_id_key;
+
 /* OS-specific initialization */
 
 static DWORD st_initialize(void)
 {
+  DWORD result = 0;
   st_thread_sem_key = TlsAlloc();
   if (st_thread_sem_key == TLS_OUT_OF_INDEXES)
     return GetLastError();
-  else
-    return 0;
+  st_thread_id_key = TlsAlloc();
+  if (st_thread_id_key == TLS_OUT_OF_INDEXES) {
+    result = GetLastError();
+    TlsFree(st_thread_sem_key);
+  }
+  return result;
 }
 
 /* Thread creation.  Created in detached mode if [res] is NULL. */
@@ -112,16 +129,32 @@ static DWORD st_tls_newkey(st_tlskey * res)
     return 0;
 }
 
-static INLINE void * st_tls_get(st_tlskey k)
+Caml_inline void * st_tls_get(st_tlskey k)
 {
   return TlsGetValue(k);
 }
 
-static INLINE void st_tls_set(st_tlskey k, void * v)
+Caml_inline void st_tls_set(st_tlskey k, void * v)
 {
   TlsSetValue(k, v);
 }
 
+/* OS-specific handling of the OCaml thread ID (must be called with the runtime
+   lock). */
+Caml_inline void st_thread_set_id(intnat id)
+{
+  CAMLassert(id != 0);
+  st_tls_set(st_thread_id_key, (void *)id);
+}
+
+/* Return the identifier for the current thread. The 0 value is reserved. */
+Caml_inline intnat st_current_thread_id(void)
+{
+  intnat id = (intnat)st_tls_get(st_thread_id_key);
+  CAMLassert(id != 0);
+  return id;
+}
+
 /* The master lock.  */
 
 typedef CRITICAL_SECTION st_masterlock;
@@ -133,27 +166,27 @@ static void st_masterlock_init(st_masterlock * m)
   EnterCriticalSection(m);
 }
 
-static INLINE void st_masterlock_acquire(st_masterlock * m)
+Caml_inline void st_masterlock_acquire(st_masterlock * m)
 {
   TRACE("st_masterlock_acquire");
   EnterCriticalSection(m);
   TRACE("st_masterlock_acquire (done)");
 }
 
-static INLINE void st_masterlock_release(st_masterlock * m)
+Caml_inline void st_masterlock_release(st_masterlock * m)
 {
   LeaveCriticalSection(m);
   TRACE("st_masterlock_released");
 }
 
-static INLINE int st_masterlock_waiters(st_masterlock * m)
+Caml_inline int st_masterlock_waiters(st_masterlock * m)
 {
   return 1;                     /* info not maintained */
 }
 
 /* Scheduling hints */
 
-static INLINE void st_thread_yield(st_masterlock * m)
+Caml_inline void st_thread_yield(st_masterlock * m)
 {
   LeaveCriticalSection(m);
   Sleep(0);
@@ -162,53 +195,106 @@ static INLINE void st_thread_yield(st_masterlock * m)
 
 /* Mutexes */
 
-typedef CRITICAL_SECTION * st_mutex;
+struct st_mutex_ {
+  CRITICAL_SECTION crit;
+  volatile st_tid owner;    /* 0 if unlocked */
+  /* The "owner" field is not always protected by "crit"; it is also
+     accessed without holding "crit", using the Interlocked API for
+     atomic accesses */
+};
+
+typedef struct st_mutex_ * st_mutex;
 
 static DWORD st_mutex_create(st_mutex * res)
 {
-  st_mutex m = caml_stat_alloc_noexc(sizeof(CRITICAL_SECTION));
+  st_mutex m = caml_stat_alloc_noexc(sizeof(struct st_mutex_));
   if (m == NULL) return ERROR_NOT_ENOUGH_MEMORY;
-  InitializeCriticalSection(m);
+  InitializeCriticalSection(&m->crit);
+  m->owner = 0;
   *res = m;
   return 0;
 }
 
 static DWORD st_mutex_destroy(st_mutex m)
 {
-  DeleteCriticalSection(m);
+  DeleteCriticalSection(&m->crit);
   caml_stat_free(m);
   return 0;
 }
 
-static INLINE DWORD st_mutex_lock(st_mutex m)
+/* Unused under Windows because there is no fork() */
+static DWORD st_mutex_init(st_mutex m)
 {
-  TRACE1("st_mutex_lock", m);
-  EnterCriticalSection(m);
-  TRACE1("st_mutex_lock (done)", m);
   return 0;
 }
 
 /* Error codes with the 29th bit set are reserved for the application */
 
-#define PREVIOUSLY_UNLOCKED 0
-#define ALREADY_LOCKED (1<<29)
+#define MUTEX_DEADLOCK (1<<29 | 1)
+#define MUTEX_PREVIOUSLY_UNLOCKED 0
+#define MUTEX_ALREADY_LOCKED (1 << 29)
+#define MUTEX_NOT_OWNED (1<<29 | 2)
 
-static INLINE DWORD st_mutex_trylock(st_mutex m)
+Caml_inline DWORD st_mutex_lock(st_mutex m)
 {
-  TRACE1("st_mutex_trylock", m);
-  if (TryEnterCriticalSection(m)) {
-    TRACE1("st_mutex_trylock (success)", m);
-    return PREVIOUSLY_UNLOCKED;
-  } else {
-    TRACE1("st_mutex_trylock (failure)", m);
-    return ALREADY_LOCKED;
+  st_tid self, prev;
+  TRACE1("st_mutex_lock", m);
+  self = st_current_thread_id();
+  /* Critical sections are recursive locks, so this will succeed
+     if we already own the lock */
+  EnterCriticalSection(&m->crit);
+  /* Record that we are the owner of the lock */
+  prev = Tid_Atomic_Exchange(&m->owner, self);
+  if (prev != 0) {
+    /* The mutex was already locked by ourselves.
+       Cancel the EnterCriticalSection above and return an error. */
+    TRACE1("st_mutex_lock (deadlock)", m);
+    LeaveCriticalSection(&m->crit);
+    return MUTEX_DEADLOCK;
   }
+  TRACE1("st_mutex_lock (done)", m);
+  return 0;
 }
 
-static INLINE DWORD st_mutex_unlock(st_mutex m)
+Caml_inline DWORD st_mutex_trylock(st_mutex m)
 {
+  st_tid self, prev;
+  TRACE1("st_mutex_trylock", m);
+  self = st_current_thread_id();
+  if (! TryEnterCriticalSection(&m->crit)) {
+    TRACE1("st_mutex_trylock (failure)", m);
+    return MUTEX_ALREADY_LOCKED;
+  }
+  /* Record that we are the owner of the lock */
+  prev = Tid_Atomic_Exchange(&m->owner, self);
+  if (prev != 0) {
+    /* The mutex was already locked by ourselves.
+       Cancel the EnterCriticalSection above and return "already locked". */
+    TRACE1("st_mutex_trylock (already locked by self)", m);
+    LeaveCriticalSection(&m->crit);
+    return MUTEX_ALREADY_LOCKED;
+  }
+  TRACE1("st_mutex_trylock (done)", m);
+  return MUTEX_PREVIOUSLY_UNLOCKED;
+}
+
+Caml_inline DWORD st_mutex_unlock(st_mutex m)
+{
+  st_tid self, prev;
+  /* If the calling thread holds the lock, m->owner is stable and equal
+     to st_current_thread_id().
+     Otherwise, the value of m->owner can be 0 (if the mutex is unlocked)
+     or some other thread ID (if the mutex is held by another thread),
+     but is never equal to st_current_thread_id(). */
+  self = st_current_thread_id();
+  prev = Tid_Atomic_Compare_Exchange(&m->owner, 0, self);
+  if (prev != self) {
+    /* The value of m->owner is unchanged */
+    TRACE1("st_mutex_unlock (error)", m);
+    return MUTEX_NOT_OWNED;
+  }
   TRACE1("st_mutex_unlock", m);
-  LeaveCriticalSection(m);
+  LeaveCriticalSection(&m->crit);
   return 0;
 }
 
@@ -291,6 +377,8 @@ static DWORD st_condvar_wait(st_condvar c, st_mutex m)
 {
   HANDLE ev;
   struct st_wait_list wait;
+  DWORD rc;
+  st_tid self, prev;
 
   TRACE1("st_condvar_wait", c);
   /* Recover (or create) the event associated with the calling thread */
@@ -303,14 +391,23 @@ static DWORD st_condvar_wait(st_condvar c, st_mutex m)
     if (ev == NULL) return GetLastError();
     TlsSetValue(st_thread_sem_key, (void *) ev);
   }
-  EnterCriticalSection(&c->lock);
+  /* Get ready to release the mutex */
+  self = st_current_thread_id();
+  prev = Tid_Atomic_Compare_Exchange(&m->owner, 0, self);
+  if (prev != self) {
+    /* The value of m->owner is unchanged */
+    TRACE1("st_condvar_wait: error: mutex not held", m);
+    return MUTEX_NOT_OWNED;
+  }
   /* Insert the current thread in the waiting list (atomically) */
+  EnterCriticalSection(&c->lock);
   wait.event = ev;
   wait.next = c->waiters;
   c->waiters = &wait;
   LeaveCriticalSection(&c->lock);
-  /* Release the mutex m */
-  LeaveCriticalSection(m);
+  /* Finish releasing the mutex m (like st_mutex_unlock does, minus
+     the error checking, which we've already done above). */
+  LeaveCriticalSection(&m->crit);
   /* Wait for our event to be signaled.  There is no risk of lost
      wakeup, since we inserted ourselves on the waiting list of c
      before releasing m */
@@ -318,9 +415,10 @@ static DWORD st_condvar_wait(st_condvar c, st_mutex m)
   if (WaitForSingleObject(ev, INFINITE) == WAIT_FAILED)
     return GetLastError();
   /* Reacquire the mutex m */
-  TRACE1("st_condvar_wait: restarted, acquiring mutex", m);
-  EnterCriticalSection(m);
-  TRACE1("st_condvar_wait: acquired mutex", m);
+  TRACE1("st_condvar_wait: restarted, acquiring mutex", c);
+  rc = st_mutex_lock(m);
+  if (rc != 0) return rc;
+  TRACE1("st_condvar_wait: acquired mutex", c);
   return 0;
 }
 
@@ -375,16 +473,28 @@ static void st_check_error(DWORD retcode, char * msg)
 
   if (retcode == 0) return;
   if (retcode == ERROR_NOT_ENOUGH_MEMORY) caml_raise_out_of_memory();
-  ret = FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM|FORMAT_MESSAGE_IGNORE_INSERTS,
-                      NULL,
-                      retcode,
-                      0,
-                      err,
-                      sizeof(err)/sizeof(wchar_t),
-                      NULL);
-  if (! ret) {
-    ret =
-      swprintf(err, sizeof(err)/sizeof(wchar_t), L"error code %lx", retcode);
+  switch (retcode) {
+  case MUTEX_DEADLOCK:
+    ret = swprintf(err, sizeof(err)/sizeof(wchar_t),
+                   L"Mutex is already locked by calling thread");
+    break;
+  case MUTEX_NOT_OWNED:
+    ret = swprintf(err, sizeof(err)/sizeof(wchar_t),
+                   L"Mutex is not locked by calling thread");
+    break;
+  default:
+    ret = FormatMessage(
+             FORMAT_MESSAGE_FROM_SYSTEM|FORMAT_MESSAGE_IGNORE_INSERTS,
+             NULL,
+             retcode,
+             0,
+             err,
+             sizeof(err)/sizeof(wchar_t),
+             NULL);
+    if (! ret) {
+      ret =
+        swprintf(err, sizeof(err)/sizeof(wchar_t), L"error code %lx", retcode);
+    }
   }
   msglen = strlen(msg);
   errlen = win_wide_char_to_multi_byte(err, ret, NULL, 0);
@@ -412,13 +522,6 @@ static DWORD WINAPI caml_thread_tick(void * arg)
   return 0;
 }
 
-/* "At fork" processing -- none under Win32 */
-
-static DWORD st_atfork(void (*fn)(void))
-{
-  return 0;
-}
-
 /* Signal handling -- none under Win32 */
 
 value caml_thread_sigmask(value cmd, value sigs) /* ML */
