@@ -41,7 +41,8 @@ module type IO_FILE = sig
   (** Write some contents to a file *)
   val write: t typed_file -> t -> unit
 
-  (** Read file contents. Raise an error if the file does not exist. *)
+  (** Read file contents.
+      @raise OpamSystem.Internal_error if the file does not exist. *)
   val read: t typed_file -> t
 
   (** Returns [None] on non-existing file *)
@@ -60,7 +61,7 @@ module type IO_FILE = sig
 
 end
 
-(* Error less [IO_FILE] read functions. *)
+(* Error less {!IO_FILE} read functions. *)
 module type BestEffortRead = sig
   type t
   val read: t typed_file -> t
@@ -584,22 +585,18 @@ module OPAM: sig
       This can be used to locate e.g. the files/ overlays *)
   val metadata_dir: t -> (repository_name option * string) option
 
-  (** Gets the resolved metadata dir, given a mapping of repository names to
-      their roots *)
-  val get_metadata_dir:
-    repos_roots:(repository_name -> dirname) -> t -> dirname option
-
   (** Names and hashes of the files below files/ *)
   val extra_files: t -> (OpamFilename.Base.t * OpamHash.t) list option
 
   (** From locked opam file *)
   val locked: t -> string option
 
-  (** Looks up the extra files, and returns their full paths, relative path to
-      the package source, and hash. Doesn't check the hashes. *)
+  (** Looks up the extra files, and returns their relative path to
+      the package source, content if it exists and hash.
+      Doesn't check the hashes. *)
   val get_extra_files:
-    repos_roots:(repository_name -> dirname) ->
-    t -> (filename * basename * OpamHash.t) list
+    get_repo_files:(repository_name -> string -> (basename * string Lazy.t) list) ->
+    t -> (basename * string Lazy.t option * OpamHash.t) list
 
   (** Returns the errors that were found when parsing the file, associated to
       their fields (that were consequently ignored) *)
@@ -1028,7 +1025,7 @@ module Repo_config_legacy : sig
 end
 
 module Repos_config: sig
-  type t = (url * trust_anchors option) option OpamRepositoryName.Map.t
+  type t = (url * trust_anchors option) OpamRepositoryName.Map.t
   include IO_FILE with type t := t
   module BestEffort: BestEffortRead with type t := t
 end
@@ -1128,13 +1125,6 @@ end
 (** {2 urls.txt file *} *)
 module File_attributes: IO_FILE with type t = file_attribute_set
 
-module Stats: sig
-
-  (** Display statistics about file access. *)
-  val print: unit -> unit
-
-end
-
 (** Helper module for manipulation of the raw syntax ([opamfile]) format.
     (the specific file handling modules are derived from this one) *)
 module Syntax : sig
@@ -1162,7 +1152,6 @@ end
 
 module type SyntaxFileArg = sig
   val internal: string
-  val atomic: bool
   val format_version: OpamVersion.t
   type t
   val empty: t
@@ -1173,7 +1162,6 @@ module SyntaxFile(X: SyntaxFileArg) : IO_FILE with type t := X.t
 
 module type LineFileArg = sig
   val internal: string
-  val atomic: bool
   type t
   val empty: t
   val pp: (string list list, t) OpamPp.t
