@@ -1,15 +1,15 @@
 (* TEST
 include tool-ocaml-lib
-flags = "-w a"
+flags = "-w -a"
 ocaml_script_as_argument = "true"
 * setup-ocaml-build-env
 ** ocaml
 *)
 
 open Lib;;
-if Pervasives.int_of_string "123" <> 123 then raise Not_found;;
+if Stdlib.int_of_string "123" <> 123 then raise Not_found;;
 (** test for fix of bug 6649: http://caml.inria.fr/mantis/view.php?id=6649 *)
-if Pervasives.int_of_string "+123" <> 123 then raise Not_found;;
+if Stdlib.int_of_string "+123" <> 123 then raise Not_found;;
 
 if Int32.of_string "+123" <> Int32.of_int 123 then raise Not_found;;
 if Int64.of_string "+123" <> Int64.of_int 123 then raise Not_found;;
