@@ -1,5 +1,5 @@
 (* TEST
-flags = " -w a "
+flags = " -w -a "
 * setup-ocamlc.byte-build-env
 ** ocamlc.byte
 *** check-ocamlc.byte-output
@@ -19,11 +19,12 @@ module type Analysis = sig
     [< `Location of t
     |  `Value of t
     |  `None ] as 'a
-  val of_var : ?f:string -> string -> [ `Location of _ | `Value of _  | `None ] maybe_region
+  val of_var :
+    ?f:string -> string ->
+      [ `Location of _ | `Value of _  | `None ] maybe_region
 end
 
 module Make (Analysis : Analysis) = struct
   include Analysis
   let of_var  = of_var ~f:""
 end
-
