@@ -13,11 +13,10 @@ module type S = sig
   class base : 'e -> ['e] t
 end;;
 [%%expect{|
-Line _, characters 2-27:
-    class base : 'e -> ['e] t
-    ^^^^^^^^^^^^^^^^^^^^^^^^^
+Line 2, characters 2-27:
+2 |   class base : 'e -> ['e] t
+      ^^^^^^^^^^^^^^^^^^^^^^^^^
 Error: Some type variables are unbound in this type:
          class base : 'e -> ['e] t
-       The method update has type 'e -> < update : 'a; .. > as 'a where 'e
-       is unbound
+       The method update has type 'e -> #base where 'e is unbound
 |}];;
